/**
 * Copyright 2019 - MagabeLab (Tanzania). All Rights Reserved.
 * Author Edwin Magabe    edyma50@yahoo.com
 */

import 'package:petitparser/petitparser.dart';
import 'package:simple_smart_scraper/simple_smart_scraper.dart';

/*
Below are step by step instructions of how to write your first data scraper using simple_smart_scraper package.

Before starting our tutorials, let's first define Data Scraping.
According to www.wikipedia.com , Data scraping ia a technique in which a computer program extract data from human-readable output coming from another program.

Tutorials
---------

First locate your target data to scrape , Our target data is the string below.

*/

final html_data = '''
 <h2>MagabeLab</h2>
 <h1><i>Welcome</i></h1>
 <tbody>
 <td><a href="link1" color="black">Home</a></td>
 <td><a href="link2" color="white">Info</a></td>
 </tbody>
''';

//For remote data such as html pages that need to be cleaned first, use getCleanedHtml function
//that download and provides options for cleaning the downloaded html data.
/*var cleanedHtm = await getCleanedHtml('data-url', keepTags: const <String>[
    'a',
    'td',
    'h1',
    'h2',
    'h3',
    'table',
    'body',
    'tr'
  ]);*/

/*
Tutorial 1: Parsers
------------------------

A parser class that provide the parsers to parse our data
*/
class MyParser with ParserMixin {
  //A parser to parse h2 tag  //<h2>MagabeLab</h2>
  //Use element function and provide a tag name as follows
  Parser titleParser() => element('h2');

  //A parser to parse Welcome text //<h1><i>Welcome</i></h1>
  //Here i use parentElement function and pass h1 as a tag
  Parser welcomeParser() => parentElement('h1', element('i'));

  //A parser to parse <td><a href="xxx">xxx</a></td>
  Parser tdParser() => parentElement('td', element('a'));

  //A parser to parse a tbody tag from the data, tbody has 2 child element both td elements, so i use repeat
  Parser tbodyParser() => parentElement('tbody', repeat(tdParser(), 2));

  //Return MagabeLab text from h2 tags
  String getTitle(String input) {
    String h2 = getParserResult(parser: titleParser(), input: input);
    return getElementText(tag: 'h2', input: h2);
  }

  //Return Welcome text from i tags
  String getWelcomeText(String input) {
    String h1 = getParserResult(parser: welcomeParser(), input: input);
    String i = getParserResult(parser: element('i'), input: h1);
    return getElementText(tag: 'i', input: i);
  }

  String tbody(String input) =>
      getParserResult(parser: tbodyParser(), input: input);

  Map<String, String> getLinkAndName(String input) {
    Map<String, String> map = <String, String>{};
    var tdList = getParserResults(parser: tdParser(), input: tbody(input));
    for (var td in tdList) {
      var a = getParserResult(parser: element('a'), input: td);
      var name = getElementText(tag: 'a', input: a);
      var href = getAttributeValue(tag: 'a', attribute: 'href', input: a);
      if (name.isNotEmpty && href.isNotEmpty) {
        map[name] = href;
      }
    }
    return map;
  }

  //just for fun ,using fast-lazy method
  String getLinkBasedColorAttribute(String color, String input) {
    var aList = getParserResults(parser: element('a'), input: input);
    for (var a in aList) {
      var c = getAttributeValue(tag: 'a', attribute: 'color', input: a);
      bool match = nonCaseSensitiveChars(c)
          .accept(color.trim()); //notice this match call
      if (match) {
        return getAttributeValue(tag: 'a', attribute: 'href', input: a);
      }
    }
    return '';
  }
}
/*
Now we are ready to try our parsers


main() {
  MyParser _myParser = MyParser();

  print(_myParser.getTitle(html_data)); //MagabeLab

  print(_myParser.getWelcomeText(html_data)); //Welcome

  print(_myParser.tbody(html_data)); //<tbody>
  //<td><a href="link1">Home</a></td>
  //<td><a href="link3" color="white">Info</a></td>
  //</tbody>

  print(_myParser.getLinkAndName(html_data)); //{Home: link1, Info: link2}

  print(_myParser.getLinkBasedColorAttribute('black', html_data)); //link1
  print(_myParser.getLinkBasedColorAttribute('white', html_data)); //link2

  //Event these calls, still return appropriate results because of our awesome match algorithm
  print(_myParser.getLinkBasedColorAttribute('WHITE', html_data)); //link2
  print(_myParser.getLinkBasedColorAttribute('whIte', html_data)); //link2
  print(_myParser.getLinkBasedColorAttribute('Black', html_data)); //link1
}*/

/*
Is it possible to write some kind of a decoder that will decode html page data to dart object on the fly ?
The Answer is YES, let's see the following tutorial

Tutorial 2: Decoder
----------------------
*/

/*
A decoder that decode our data and return Link model with data (href,color, name)
The data is obtained from an 'a' element that looks like  <a href="link2" color="white">Info</a>

Introducing our model class */

class Link {
  String href;
  String color;
  String name;

  Link(this.href, this.color, this.name);

  @override
  String toString() {
    return '{href:$href, color:$color name:$name}';
  }
}

//The  decoder class

class MyDecoder extends Decoder<Link> {
  @override
  Link mapParserResult(String result) {
    var name = getElementText(tag: 'a', input: result);
    var href = getAttributeValue(tag: 'a', attribute: 'href', input: result);
    var color = getAttributeValue(tag: 'a', attribute: 'color', input: result);
    if (name.isNotEmpty && href.isNotEmpty && color.isNotEmpty) {
      return Link(href, color, name);
    }
    return null;
  }

  @override
  Parser get parser => element('a');
}

//Now our decoder is ready to use and there are many ways to use it, here i only show two
/*
main() {

  MyDecoder _mydecoder = MyDecoder();

  //       Method 1: call the decode method
  //         --------------------

  _mydecoder.decode(html_data).listen((link) {
    print(link);
  }, onDone: () {
    print('done! Method 1');
  });

  /*
  The above call chain print the following

  {href:link1, color:black name:Home}
  {href:link2, color:white name:Info}
  done! Method 1

  * */

  //       Method 2: passing the decode toa stream transform function
  //         --------------------
  //First our toStream function , you don't need to write your own, the library already provided  you with
  //[stringToStream] function that you can use.

  Stream<String> toStream(txt) async* {
    yield txt;
  }

  //now we are ready for method 2
  toStream(html_data).transform(_mydecoder).expand((i) => i).listen((link) {
    print(link);
  }, onDone: () {
    print('done! Method 2');
  });

  /*
  The above call chain print the following

  {href:link1, color:black name:Home}
  {href:link2, color:white name:Info}
  done! Method 2

  * */
}*/

/*
Tutorial 3: DecoderBloc
-----------------------------

simple_smart_scraper comes with DecoderBloc that help you integrate your data scraping logic
into flutter applications, Bloc creation is simple let's see how.

A Bloc (Business Logic Component) is like a pipe, Events goes in and States come out.

First, we create data structures to represent bloc events ,
    here i am using an enum ,you can use a class to represent bloc events if you want to.
*/

enum MyBlocEvent { title, welcome, link, done }

//Second, we create data structures to represent bloc states.

class MyBlocState {}

//NOTE: Use libraries like equatable to avoid overriding  operator == and hashCode methods.
class TitleState extends MyBlocState {
  String title;
  TitleState(this.title);

  @override
  bool operator ==(Object other) =>
      identical(this, other) ||
      other is TitleState &&
          runtimeType == other.runtimeType &&
          title == other.title;

  @override
  int get hashCode => title.hashCode;
}

class WelcomeState extends MyBlocState {
  String welcome;
  WelcomeState(this.welcome);

  @override
  bool operator ==(Object other) =>
      identical(this, other) ||
      other is WelcomeState &&
          runtimeType == other.runtimeType &&
          welcome == other.welcome;

  @override
  int get hashCode => welcome.hashCode;
}

class LinkState extends MyBlocState {
  Link link;
  LinkState(this.link);

  @override
  bool operator ==(Object other) =>
      identical(this, other) ||
      other is LinkState &&
          runtimeType == other.runtimeType &&
          link == other.link;

  @override
  int get hashCode => link.hashCode;
}

class CompletedState extends MyBlocState {}

//Now we are ready to implement our bloc.

class MyBloc extends DecoderBloc<MyBlocEvent, MyBlocState> {
  MyDecoder _decoder = MyDecoder();
  MyParser _myParser = MyParser();
  String _welcome = '';
  String _title = '';
  List<Link> links = <Link>[];

  @override
  Future<void> load(String input, {String baseUrl}) {
    _welcome = '';
    _title = '';
    links.clear();
    return super.load(input, baseUrl: baseUrl);
  }

  @override
  void dispatchEvents(String input, {String baseUrl}) {
    _welcome = _myParser.getWelcomeText(input);
    if (_welcome.isNotEmpty) {
      dispatchEvent(MyBlocEvent.welcome);
    }
    _title = _myParser.getTitle(input);
    if (_title.isNotEmpty) {
      dispatchEvent(MyBlocEvent.title);
    }

    decode(
      input: input,
      decoder: _decoder,
      listener: links.add,
      onDone: () {
        dispatchEvent(MyBlocEvent.link);
        dispatchDelayedEvent(Duration(seconds: 1), MyBlocEvent.done);
      },
    );
  }

  @override
  Stream<MyBlocState> mapEventToState(MyBlocEvent event) async* {
    switch (event) {
      case MyBlocEvent.welcome:
        if (_welcome.isNotEmpty) {
          yield WelcomeState(_welcome);
        }
        break;
      case MyBlocEvent.title:
        if (_title.isNotEmpty) {
          yield TitleState(_title);
        }
        break;
      case MyBlocEvent.link:
        for (var link in links) {
          yield LinkState(link);
        }
        break;
      case MyBlocEvent.done:
        yield CompletedState();
        complete(); //must call complete() when done for the bloc to be able to reload or load new input
        break;
    }
  }
}

//Now that our bloc is ready, let's use it.

main() async {
  /* MyBloc mybloc = MyBloc();

  mybloc.listen((state) {
    print('State = $state');
  });

  mybloc.load(html_data);*/

  /*
      The above print the following


      State = Instance of 'WelcomeState'
      State = Instance of 'TitleState'
      State = Instance of 'LinkState'
      State = Instance of 'LinkState'
      State = Instance of 'CompletedState'

  * */

  var block = MyBloc()
    ..listen((state) {
      if (state is WelcomeState) {
        print(state.welcome);
      } else if (state is TitleState) {
        print(state.title);
      } else if (state is LinkState) {
        print(state.link);
      } else if (state is CompletedState) {
        print('MyBloc completed');
      }
    });
  await block.load(html_data);

  /*
      The above print the following


        Welcome
        MagabeLab
        {href:link1, color:black name:Home}
        {href:link2, color:white name:Info}
        MyBloc completed

   */
}
